/*
  File name        : core.c
  Date of creation : 10/2/2008
  Version          : 1.0
  Copyright        : Soft'n'Design
  Author           : Laurent Mazet <mazet@softndesign.org>

  Description      : Miscellaenous functions on core/thread management

  History          :
  - initial version from tools.h
*/

#define _GNU_SOURCE
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>

#include "rdtsc.h"
#include "verbose.h"

#include "core.h"

DECLARE_VERBOSE_LEVEL (coretools, 0);

int get_cpu_number (void)
{
    int ncpu = 0;
    char *procinfo = "processor";

    FILE *F = fopen ("/proc/cpuinfo", "r");
    if (F) {
        char line[256] = { '\0' };
        while (fgets (line, sizeof (line) - 1, F)) {
            if (strncmp (line, procinfo, strlen (procinfo)) == 0)
                ncpu++;
        }
        fclose (F);
    }

    return (ncpu == 0) ? 1 : ncpu;
}


/* Thread management */

int enable_realtime (int prio)
{
    struct sched_param param;

    param.__sched_priority = prio;
    return pthread_setschedparam (pthread_self (), SCHED_FIFO, &param);
}

int cpu_setaffinity (const char *cpu_list)
{
    cpu_set_t cpuset;
    char *save_ptr;
    char *cpu;
    char delim[] = "+, ";

    if (!cpu_list) return -1;

    char *copy_list = strdup(cpu_list);

    CPU_ZERO(&cpuset);
    cpu = strtok_r(copy_list, delim, &save_ptr);
    while (cpu) {
        CPU_SET(atoi(cpu), &cpuset);
        cpu = strtok_r(NULL, delim, &save_ptr);
    }

    free(copy_list);

    return pthread_setaffinity_np(pthread_self (), sizeof(cpu_set_t), &cpuset);
}

/**
   Cpu clock
*/
static double cpu_clock = 1;


/**
    Internal funciton, executed at load time to calibrate cpu clock
*/
void __attribute__((constructor)) _core_init_cpu_clock (void)
{
    struct timeval tv1, tv2;
    long long int ts1, ts2;

    gettimeofday (&tv1, NULL);
    ts1 = rdtsc ();

    usleep (100000);

    gettimeofday (&tv2, NULL);
    ts2 = rdtsc ();

    double delta = ((double)tv2.tv_sec + tv2.tv_usec * 1e-6) -
        ((double)tv1.tv_sec + tv1.tv_usec * 1e-6);

    cpu_clock = (ts2 - ts1) / delta;
}

unsigned long long tsc2diff (unsigned long long tsc1, unsigned long long tsc2)
{
    return (tsc1 > tsc2) ? tsc1 - tsc2 : tsc2 + (~tsc1) + 1;
}

float tsc2ms (unsigned long long delta_tsc)
{
    return (float) (delta_tsc / cpu_clock * 1e6f);
}

/* vi:set tabstop=4 expandtab shiftwidth=4: this line set vi mode */
